import glob
import random
from skimage import io, transform
from torch.utils import data
from torchvision.transforms import ToTensor
from torchvision.transforms.functional import resize
import warnings

class PolyUDataset(data.Dataset):
    DATASET_BASE_DIR = '../../data/PolyU-Real-World-Noisy-Images-Dataset/'
    DATASET_SEED = '84239482109'
    DATA_SPLIT_PERCENTS = {'train' : (0.0, 0.66),
                           'val' : (0.66, 0.83),
                           'test' : (0.83, 1.0),
                           'all' : (0.0, 1.0)}
                           
    def __init__(self, split_type='train', use_cropped_images=True, image_type='real',
                 in_memory=True, dataset_dir=DATASET_BASE_DIR, downsample_shape=None,
                 as_grayscale=False):
        """
            Args:
                split_type: Whether to retrieve training, validation, or testing data.
                             Possible values: ['train', 'test', 'val', 'all']
                use_cropped_images: If True, retrieves the images cropped to 512 x 512.
                                     If False, all images will have their original size.
                image_type: Whether to retrieve only the "mean", "real", or all images
                             "real" images contain a good amount of noise
                             "mean" images are "real" images except smoothed, probably
                              with an average filter
                             Possible values: ['mean', 'real', 'all']
                in_memory: Whether or not to cache all images in memory 
                dataset_dir: The file path to the directory with the CroppedImages/ and
                              OriginalImages/ subfolders
                downsample_shape: The size tuple to resize the loaded images to. If None
                                   no reshaping will be done on the loaded dataset
                as_grayscale: If True, images will be converted to grayscale
        """
        assert(split_type in ['train', 'test', 'val', 'all'])
        assert(image_type in ['mean', 'real', 'all'])
        if use_cropped_images:
            dataset_dir += 'CroppedImages/'
        else:
            dataset_dir += 'OriginalImages/'
        self.in_memory = in_memory
        self.downsample_shape = downsample_shape
        self.as_grayscale = as_grayscale
        
        # Find only the images labelled with the suffix specified by image_type
        image_file_path_glob = dataset_dir + '*.[jJ][pP][gG]'
        if image_type != 'all':
            image_file_path_glob = dataset_dir + '*_' + image_type + '.[jJ][pP][gG]'        
        
        # Deterministically shuffle the data filepaths to randomly assign images to train, test, or val
        all_image_file_paths = glob.glob(image_file_path_glob)
        random.seed(PolyUDataset.DATASET_SEED)
        random.shuffle(all_image_file_paths)
        
        if len(all_image_file_paths) <= 0:
            warnings.warn(f'No image files found in dataset_dir')
                
        # Partition the data according the the split percents defined at the top of this class
        (data_start_percent, data_end_percent) = PolyUDataset.DATA_SPLIT_PERCENTS[split_type]
        split_start_index = int(data_start_percent * len(all_image_file_paths))
        split_end_index = int(data_end_percent * len(all_image_file_paths))
        self.split_data_file_paths = all_image_file_paths[split_start_index:split_end_index]
        
        # If in_memory is specified, cache all images for the split in memory
        if self.in_memory:
            self.image_data = [io.imread(image_file_path, as_gray=as_grayscale)
                               for image_file_path
                               in self.split_data_file_paths]
            if self.downsample_shape is not None:
                self.image_data = [transform.resize(img, self.downsample_shape) for img in self.image_data]
                    
        
    def __len__(self):
        return len(self.split_data_file_paths)
        
    def __getitem__(self, index):
        if self.in_memory:
            index_image = self.image_data[index]
        else:
            image_file_path = self.split_data_file_paths[index]
            index_image = io.imread(image_file_path, as_gray=as_grayscale)
            
            if self.downsample_shape is not None:
                index_image = transform.resize(index_image, self.downsample_shape)
            
        return ToTensor()(index_image)


if __name__ == "__main__":
    TRAIN_BATCH_SIZE = 8
    VAL_BATCH_SIZE = 8
    TEST_BATCH_SIZE = 16
    IMAGE_TYPE = 'mean' # can also be 'real' or 'all'

    # Example dataloaders for training, validation, and testing
    training_dataset = PolyUDataset(split_type='train', image_type=IMAGE_TYPE)
    training_dataloader = data.DataLoader(training_dataset, batch_size=TRAIN_BATCH_SIZE, shuffle=True, num_workers=0, drop_last=True)
    print(f'Training: {len(training_dataset)} images')
    for (i, current_image_batch) in enumerate(training_dataloader):
        print(f'Batch {i}: {current_image_batch.shape}')
    print()

    validation_dataset = PolyUDataset(split_type='val', image_type=IMAGE_TYPE, in_memory=False)
    validation_dataloader = data.DataLoader(validation_dataset, batch_size=VAL_BATCH_SIZE, shuffle=False, num_workers=0, drop_last=False)
    print(f'Validation: {len(validation_dataset)} images')
    for (i, current_image_batch) in enumerate(validation_dataloader):
        print(f'Batch {i}: {current_image_batch.shape}')
    print()

    test_dataset = PolyUDataset(split_type='test', image_type=IMAGE_TYPE)
    test_dataloader = data.DataLoader(test_dataset, batch_size=TEST_BATCH_SIZE, shuffle=False, num_workers=0, drop_last=False)
    print(f'Testing: {len(test_dataset)} images')
    for (i, current_image_batch) in enumerate(test_dataloader):
        print(f'Batch {i}: {current_image_batch.shape}')
    print()


import torch 
import torch.nn as nn

def torch_kron(A, B):
    return torch.einsum('ij, kl->ikjl', A, B).view((A.shape[0] * B.shape[0], A.shape[1] * B.shape[1]))

class DirectTuckerNet(nn.Module):
    def __init__(self, target_dims, latent_dims, nonnegative=False):
        super(DirectTuckerNet, self).__init__()
        self.A = nn.Parameter(torch.Tensor(target_dims[0], latent_dims[0]))
        self.B = nn.Parameter(torch.Tensor(target_dims[1], latent_dims[1]))
        self.C = nn.Parameter(torch.Tensor(target_dims[2], latent_dims[2]))

        self.G = nn.Parameter(torch.Tensor(latent_dims[0], latent_dims[1] * latent_dims[2]))
        
        nn.init.uniform_(self.A)
        nn.init.uniform_(self.B)
        nn.init.uniform_(self.C)
        nn.init.uniform_(self.G)

        if nonnegative:
            self.sm = nn.Softmax(-1)
            self.relu = nn.ReLU()
        else:
            self.sm = nn.Identity()
            self.relu = nn.Identity()
        
    def get_ABCG(self):
        return self.A.data, self.B.data, self.C.data, self.G.data
    
    def forward(self):
        C_KR_B = torch_kron(self.sm(self.C), self.sm(self.B))
        return self.sm(self.A) @ self.relu(self.G ) @ C_KR_B.transpose(-1, -2)


import torch 
import torch.nn as nn
import numpy as np

def torch_kron(A, B):
    return torch.einsum('ij, kl->ikjl', A, B).view((A.shape[0] * B.shape[0], A.shape[1] * B.shape[1]))


class AutoEncoderTucker(nn.Module):
    def __init__(self, target_dims, latent_dims):
        super(AutoEncoderTucker, self).__init__()    
        B = np.random.uniform(size=(target_dims[1], latent_dims[1]))
        C = np.random.uniform(size=(target_dims[2], latent_dims[2]))
        G = np.random.uniform(size=(latent_dims[0], latent_dims[1] * latent_dims[2]))
        
        self.B = nn.Parameter(torch.Tensor(B))
        self.C = nn.Parameter(torch.Tensor(C))
        self.G = nn.Parameter(torch.Tensor(G))
        
        
        self.B_inv = nn.Parameter(torch.Tensor(np.linalg.pinv(B)))
        self.C_inv = nn.Parameter(torch.Tensor(np.linalg.pinv(C)))
        self.G_inv = nn.Parameter(torch.Tensor(np.linalg.pinv(G)))
        
        self.sm = nn.Softmax(-1)
        self.relu = nn.ReLU()

    def get_A(self, X):
        return self.encode(X)
    
    def get_BCG(self):
        return self.B.data, self.C.data, self.G.data
    
    def encode(self, X):
        X_T = X.transpose(0, 1)
        C_KR_B_INV = torch_kron(self.C_inv, self.B_inv)
        G_inv_T = self.G_inv.transpose(-1, -2)
        A_T = G_inv_T @ C_KR_B_INV @ X_T
        return A_T.transpose(-1, -2)
    
    def decode(self, A):
        C_KR_B = torch_kron(self.sm(self.C), self.sm(self.B))
        
        return self.sm(A) @ self.relu(self.G) @ C_KR_B.transpose(-1, -2)
    
    def forward(self, X):
        return self.decode(self.encode(X))


import numpy as np
import torch
import torch.nn as nn

class SVDNet(nn.Module):
    """SVD implemented as a neural network in PyTorch"""

    class HouseHolderLayer(nn.Module):
        """Householder layer for SVD network"""

        def __init__(self, n, num_hhrs):
            super(SVDNet.HouseHolderLayer, self).__init__()
            
            self.hhrs = nn.ParameterList([])
            for i in range(num_hhrs):
                self.hhrs.append(nn.Parameter(torch.Tensor(n)))

        def mult_by_hhr(self, x, h):
            hT = h.unsqueeze(0)
            h = h.unsqueeze(1)
            beta = 2 / (h**2).sum()

            return x - beta * (h @ (hT @ x)) 
        
        def forward(self, x):
            res = self.mult_by_hhr(x, self.hhrs[0])
            for i in range(1, len(self.hhrs)):
                res = self.mult_by_hhr(res, self.hhrs[i])
            return res


    def __init__(self, n, num_hhrs):
        """Params
        n: TODO INSERT HERE
        num_hhrs: TODO INSERT HERE
        """
        super(SVDNet, self).__init__()
        self.sigma = nn.Parameter(torch.ones(n,1))
        
        self.U = SVDNet.HouseHolderLayer(n, num_hhrs)
        self.VT = SVDNet.HouseHolderLayer(n, num_hhrs)
        
        for p in self.U.parameters():
            torch.nn.init.normal_(p)
        for p in self.VT.parameters():
            torch.nn.init.normal_(p)
    

    def get_USVT(self, device):
        eye = torch.eye(self.sigma.shape[0]).to(device)
        U = self.U(eye).detach().cpu().numpy()
        
        S = np.diag(self.sigma.data.cpu().numpy()[:,0])

        VT = self.VT(eye).detach().cpu().numpy()
        return U, S, VT
    

    def forward(self, x):
        return self.U(self.sigma * self.VT(x))



import matplotlib.pyplot as plt
import torch

class Experiment:
    """General experiment wrapper.
    
    For running an experiment with the usual setup of 
    data as input to the network, with ground truths (labels), and calculating accuracies,
    the constructor would be something like:
    
    experiment = Experiment(net=MY_NETWORK, loss=MY_LOSS, optimizer=MY_OPTIMIZER,
        train_dataloader=MY_TRAIN_DATALOADER, validation_dataloader=MY_VALIDATION_DATALOADER,
        cuda=(True/False), accuracy_func=MY_ACCURACY_FUNC
    )
    
    train_losses, val_losses, val_accs = experiment.run(
        train_epochs=NUM_EPOCHS, train_validation_interval=VAL_INTERVAL
    )
    """
    
    REQUIRED_PARAMS = ["net", "loss", "optimizer",
                       "train_dataloader", "validation_dataloader"]
    
    def __init__(self, **kwargs):
        """Params
            net (required): PyTorch nn module to be trained and evaluated
            loss (required): loss function used in training
            optimizer (required): optimizer function used in training
            train_dataloader (required): PyTorch dataloader object used for training
            validation_dataloader (required): PyTorch dataloader object used for validation
        
            cuda (optional): use cuda, defaults to False
            use_eye_as_net_input (optional): use identity matrix as network input, defaults to False
            inputs_are_ground_truth (optional): use the input data as the ground truth
                in loss and accuracy functions
            accuracy_func (optional): functional for computing output accuracy as compared to ground truth
        """
        self._validate_required_params(kwargs)
        
        # required params
        self.net                     = kwargs["net"]
        self.loss_func               = kwargs["loss"]
        self.optimizer               = kwargs["optimizer"]
        self.train_dataloader        = kwargs["train_dataloader"]
        self.validation_dataloader   = kwargs["validation_dataloader"]
        
        # optional params
        self.cuda                    = kwargs["cuda"] if "cuda" in kwargs else False
        self.use_eye_as_net_input    = kwargs["use_eye_as_net_input"] if "use_eye_as_net_input" in kwargs else False
        self.inputs_are_ground_truth = kwargs["inputs_are_ground_truth"] if "inputs_are_ground_truth" in kwargs else False
        self.accuracy_func           = kwargs["accuracy_func"] if "accuracy_func" in kwargs else None
        self.has_labels              = kwargs["has_labels"] if "has_labels" in kwargs else True
        self.print_all_epochs        = kwargs["print_all_epochs"] if "print_all_epochs" in kwargs else False
        
        
    def run(self, train_epochs, train_validation_interval):
        """Params
            train_epochs: number of epochs to train for
            train_validation_interval: the epoch frequency which to calculate validation statistics
        
        Returns
            train_loss_over_epochs (list of floats): train loss for every epoch
            val_loss_over_epochs (list of floats): validation loss for each epoch at interval
            val_acc_over_epochs (None or list of floats): None if no accuracy function provided,
                accuracies otherwise
        """
        if train_epochs <= 0:
            raise Exception(f"Error: train_epochs must be greater than zero. {train_epochs} provided is invalid.")
            
        train_loss_over_epochs, val_loss_over_epochs, val_acc_over_epochs = self._train(
            train_epochs, train_validation_interval
        )
        
        return train_loss_over_epochs, val_loss_over_epochs, val_acc_over_epochs
        
        
    def plot_result(self, xs, ys, xlabel, ylabel, title):
        plt.plot(xs, ys)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xticks(xs)
        plt.show()
        
        
    def _train(self, num_epochs, validation_interval):    
        train_loss_over_epochs = []
        val_loss_over_epochs = []
        val_acc_over_epochs = [] if self.accuracy_func else None
        
        for epoch in range(1, num_epochs + 1):
            running_loss = 0.0
  
            for i, data  in enumerate(self.train_dataloader, 0):  
                self.optimizer.zero_grad()
            
                loss, _ = self._run_network(data, calc_accuracy=False)
                    
                loss.backward()
                self.optimizer.step()

                running_loss += loss.item()
                
            train_loss_over_epochs.append(running_loss)
            if self.print_all_epochs or epoch % validation_interval == 0:
                print(f"epoch {epoch}: total loss is {running_loss}, avg loss is {running_loss / len(self.train_dataloader)}")
                
            if epoch % validation_interval == 0:
                validation_loss, validation_acc = self._validation()
                val_loss_over_epochs.append(validation_loss)
                
                outstr = f"epoch {epoch}: val avg loss is {validation_loss}"
                if self.accuracy_func:
                    val_acc_over_epochs.append(validation_acc)
                    outstr += f". val avg acc is {validation_acc}"
                    
                print(outstr)
                
                
        return train_loss_over_epochs, val_loss_over_epochs, val_acc_over_epochs
        
    
    def _validation(self):
        self.net.eval()
        
        running_loss = 0.0
        running_acc = 0.0 if self.accuracy_func else None
        
        for i, data  in enumerate(self.validation_dataloader, 0):
            loss, acc = self._run_network(data, calc_accuracy=True)
            running_loss += loss.item()
            
            if acc:
                running_acc += acc
                   
        avg_loss = running_loss / len(self.validation_dataloader)
        
        if running_acc:
            avg_acc = running_acc / len(self.validation_dataloader)
        else:
            avg_acc = None
            
        self.net.train()
        return avg_loss, avg_acc
    
    
    def _run_network(self, data, calc_accuracy=False):
        if self.has_labels:
            inputs, gts = data
        else:
            inputs = data
                                    
        if self.cuda:
            inputs = inputs.cuda()
                    
        if self.use_eye_as_net_input:
            # dim 0 is batch size, dim 1 is number of channels (e.g., 3 for color image)
            outputs = self.net(torch.eye(inputs.shape[2]))
        else:
            outputs = self.net(inputs)

        if self.inputs_are_ground_truth:
            loss = self.loss_func(outputs, inputs)
        else:
            loss = self.loss_func(outputs, gts)
        
        if calc_accuracy and self.accuracy_func is not None:
            if self.inputs_are_ground_truth:
                acc = self.accuracy_func(outputs, inputs)
            else:
                acc = self.accuracy_func(outputs, gts)
        else:
            acc = None
        
        return loss, acc
            
        
    def _validate_required_params(self, params):
        for p in self.REQUIRED_PARAMS:
            if p not in params:
                raise Exception(f"Error: required param {p} not found in Experiment args")


import torch
import torch.nn as nn

from numpy.random import uniform
from numpy.linalg import pinv

class NMFNet(nn.Module):
    """NMF implemented as a neural network"""
    
    def __init__(self, X_height, k):
        """Params
        X_height: TODO INSERT DESC HERE
        k: TODO INSERT DESC HERE
        """
        super(NMFNet, self).__init__()
        
        self.k = k
        
        W_numpy = uniform(0, 1, (X_height, k))
        W_numpy = W_numpy / W_numpy.sum(0)[None,:]
        self.W = nn.Parameter(torch.FloatTensor(W_numpy))
        self.W_inv = nn.Parameter(torch.FloatTensor(pinv(W_numpy)))
        
        self.relu = nn.ReLU()
        self.sm = nn.Softmax(dim=0)
        
    def forward(self, X):
        H = self.get_H(X) # this is the encoder
        
        X_hat = self.sm(self.W) @ H # this is the decoder
        
        return X_hat
    
    
    def get_H(self, X):
        return self.relu(self.W_inv @ X)
    
    
    def get_W(self):
        return self.sm(self.W)


from setuptools import setup, find_packages
setup(
    name="torchfactor",
    packages=find_packages(),
    version="1.0.0",
    license="MIT",
    description="A package for tensor factorizations using neural networks in PyTorch",
    author="Jonah Casebeer, Aaron Green, Alex Mackowiack",
    author_email="jonahmc2@illinois.edu, aarongg2@illinois.edu, amackow2@illinois.edu",
    url="https://github.com/Aaron09/torchfactor",
    download_url="https://github.com/Aaron09/torchfactor/archive/v1.0.0.tar.gz",
    keywords=["PyTorch", "tensor", "factorization", "neural network", "deep learning"],
    install_requires=[
        "torch",
        "torchvision",
        "numpy",
        "matplotlib",
        "scikit-image"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3.6"
    ],
)
